# THIS WORK CONTAINS TRADE SECRET AND CONFIDENTIAL
# INFORMATION WHICH ARE THE PROPERTY OF SIEMENS INDUSTRY
# SOFTWARE INC OR ITS LICENSORS OR AFFILIATES AND IS
# SUBJECT TO LICENSE TERMS.

# CREATES EXPLICIT MATERIAL MAP OF EACH BOARD LAYER IN ODB++ WHEN GIVE PATH TO ODB++ AND PATH TO SIMCENTER FLOTHERM
# REQUIRES numpy,psutil, opencv-python

# VERSION 2403: added option to remove internal cutouts
# VERSION 2405: updated logic to better determine internal cutouts.  Added optional geometry_model.csv import for custom
#               cutout support.
# VERSION 2407: bug fixes around layer processing
# VERSION 2408: automatic length unit extraction from geometry_model.csv and used to convert material map units to meter
# VERSION 2408.1: direct material map update with length units, rather than unit conversion
import os
import pyFS as fs
import zipfile
import time
import shutil
import tempfile
import cv2
import math
import numpy as np
import sys

# USER INPUTS
deleteInternalCutouts = True  # If FALSE will keep all cutouts. Ignored when user supplied geometry_model.csv
geomTol = 1E-6  # Shouldn't need to be changed.  Tolerance to compensate for numerical precision
# SIMCENTER FLOTHERM BIN DIRECTORY
fltBin = r'C:\Program Files\Siemens\SimcenterFlotherm\2404\WinXP\bin'
debug = False
# DEFECT WHEN TRUE
matMapUnits = False  # If true will update material map DISTANCE_UNIT string, if False will convert units to meter
# MATERIAL PROPERTIES FOR BLACK (Cu) AND WHITE (FR4) PIXELS: [Kx, Ky, Kz, Cp, Rho]
bProps, wProps = [385, 385, 385, 8930, 385], [0.3, 0.3, 0.3, 1200, 880]


# END USER INPUTS


def main():
    odbPath, geomPath = arg_process()
    # odbPath = r"D:\TME\Projects\materialMapAndPowerMap\dxTopLevel\materialmap\dxtoplevel.tgz"
    # geomPath = ''
    outputFldr = os.path.abspath(os.path.join(odbPath, os.pardir))
    # CREATE TMP SOLN FOLDER
    tmpSlnFldr = tempfile.NamedTemporaryFile().name
    if not os.path.exists(tmpSlnFldr): os.mkdir(tmpSlnFldr)
    if debug: print(tmpSlnFldr)

    strt = time.time()
    # IMPORTS ODB++ TO SIMCENTER FLOTHERM AND SAVES FLOPCB FILE [FOR LAYER IMAGES] AND EXPORTS GEOMETRY_MODEL.CSV
    # SAVED TO tmpSlnFldr
    print('Importing ' + odbPath + '...')
    userDefinedGeom = False
    if len(geomPath) > 0:
        if os.path.isfile(geomPath):
            userDefinedGeom = True
        else:
            sys.exit('Unable to find geometry_model.csv')

    floPCBpath, projNm = importODB(odbPath=odbPath, slnFldr=tmpSlnFldr, userSuppliedGeomCSV=userDefinedGeom)
    now = _repTime(strtTime=strt, msg='ODB++ import')
    print('Extracting images...')
    # EXTRACT IMAGES FROM FLOPCB FILE AND PLACE IN outputFldr, RETURN PATHS TO IMAGES
    imgPaths = floPCBimages(floPCBpath=floPCBpath, outputFldr=outputFldr)
    now = _repTime(strtTime=now, msg='Image extraction')
    print('Extracting PCB stackup information...')

    # PROCESS GEOMETRY_MODEL.CSV
    if userDefinedGeom:
        geomFile = geomPath
    else:
        geomFile = os.path.join(tmpSlnFldr, 'geometry_model.csv')
        if deleteInternalCutouts:
            intCutoutNms = getInternalCutoutNms(geometryCSVPath=geomFile)
            print('   Removing ' + str(len(intCutoutNms)) + ' internal cutouts...')
            #     Remove cutouts and create new geometry_model.csv
            removeCutouts(slnFldr=tmpSlnFldr, projNm=projNm, cutoutNms=intCutoutNms)

    brdNm, xyReg, xyCutout, zStack, lngthUnit = getEDABrdLayerGeom(geometryCSVPath=geomFile, pcbNm='')
    print('length units: ' + lngthUnit)
    now = _repTime(strtTime=now, msg='Stackup extraction')
    print('Building Material Map with ' + str(len(xyReg) * len(zStack)) + ' regions...')

    # CREATE MATERIAL MAP
    version, encrypt, sigFig = 1.0, 0, 6
    distUnit = 'm'
    if matMapUnits: distUnit = lngthUnit
    matPath = os.path.join(outputFldr, projNm + '_map.csv')
    buildMaterialMap(mapPath=matPath, boardName=projNm, blkProps=bProps, whtProps=wProps, xyRegions=xyReg,
                     zRegions=zStack, imagePaths=imgPaths, version=version, sigfig=sigFig, distUnit=distUnit,
                     encrypt=encrypt)
    now = _repTime(strtTime=now, msg='Material Map creation')
    print('Material Map written to: ' + matPath)

    # REMOVE TEMP FOLDER
    if not debug: shutil.rmtree(tmpSlnFldr)
    now = _repTime(strtTime=strt, msg='Total processing')


def getInternalCutoutNms(geometryCSVPath):
    intCutoutNms = []
    bcCutouts, intCutouts = [], []
    brdNm, xyReg, xyCutout, zStack, _ = getEDABrdLayerGeom(geometryCSVPath=geometryCSVPath, pcbNm='')

    # all cutouts
    allCutouts = [[cutOut[0]] + [float(c) for c in cutOut[1:]] for cutOut in xyCutout]
    # Boundary coordinates
    xMin, xMax = min([float(p[1]) for p in xyReg]), max([float(p[2]) for p in xyReg])
    yMin, yMax = min([float(p[3]) for p in xyReg]), max([float(p[4]) for p in xyReg])
    # cutouts touching/not touching boundary
    bcCutouts = [c for c in allCutouts if
                 _cutoutIsConnected(cutoutA=['brd', xMin, xMax, yMin, yMax], cutoutB=c, bdryTest=True)]
    intCutouts = [c for c in allCutouts if c not in bcCutouts]

    if len(bcCutouts) == 0:
        intCutoutNms = [c[0] for c in allCutouts]
    else:
        for bc in bcCutouts:
            intCcen = _sortedList(cutoutA=bc, cutOuts=intCutouts)
            cutoutContact = [c for c in intCcen if _cutoutIsConnected(cutoutA=bc, cutoutB=c)]
            for cc in cutoutContact:
                intCutouts.pop(intCutouts.index(cc))
                if cc not in bcCutouts: bcCutouts.append(cc)
        intCutoutNms = [c[0] for c in intCutouts]
    return intCutoutNms


def getInternalCutoutNmsV2403(geometryCSVPath):
    brdNm, xyReg, xyCutout, zStack = getEDABrdLayerGeom(geometryCSVPath=geometryCSVPath, pcbNm='')

    xMin, xMax = min([float(p[1]) for p in xyReg]), max([float(p[2]) for p in xyReg])
    yMin, yMax = min([float(p[3]) for p in xyReg]), max([float(p[4]) for p in xyReg])

    # find interior cutouts
    intCutouts = []
    for cutOut in xyCutout:
        nM, xs, xe, ys, ye = [cutOut[0]] + [float(c) for c in cutOut[1:]]
        if abs(xMin - xs) <= geomTol or abs(xMax - xe) <= geomTol or abs(yMin - ys) <= geomTol or abs(
                yMax - ye) <= geomTol:
            pass  # boundary
        else:
            intCutouts.append(nM)
    return intCutouts


def buildMaterialMap(mapPath, boardName, blkProps, whtProps, xyRegions, zRegions, imagePaths, version, sigfig,
                     distUnit='m', encrypt=0):
    delim = ','
    blk = np.array([0, 0, 0])
    blkP, whtP = delim.join([str(b) for b in blkProps]), delim.join([str(w) for w in whtProps])
    mapFile = open(mapPath, 'w')
    if encrypt not in [0, 1]: encrypt = 0
    mapFile.write('NAME,' + boardName + '\n')
    mapFile.write('VERSION,' + str(version) + '\n')
    mapFile.write('NUMBER_OF_SF,' + str(int(sigfig)) + '\n')
    mapFile.write('ENCRYPTED,' + str(encrypt) + '\n')
    mapFile.write('DISTANCE_UNIT,' + distUnit + '\n')

    fields = ['X_CONDUCTIVITY', 'Y_CONDUCTIVITY', 'Z_CONDUCTIVITY', 'SPECIFIC_HEAT', 'DENSITY']
    # DETERMINE PHYSICAL SIZE FOR PIXEL, AND MAXIMUM PIXEL IN X,Y
    xMtrPerPx, yMtrPerPx, yPxMax, xPxMax = _pixelUnits(xyRegions=xyRegions, imgPath=imagePaths[0])
    pxAdd, cnt = 1, 0
    for i, z in enumerate(zRegions):
        zNm, zType, zS, zE = z
        imgPath = [im for im in imagePaths if im.endswith(zNm + '.PNG')][0]
        img = np.flipud(cv2.imread(imgPath))
        for xyCnt, xy in enumerate(xyRegions):
            xyNm, xS, xE, yS, yE = xy

            if ' ' in xyNm:
                xyNm = xyNm.split(' ')[-1]
            else:
                xyNm = ''
            regNm = zNm + '_' + xyNm
            regNm = regNm.strip('_')
            if debug: print(regNm)
            # GET IMAGE BLOCK AND PROPERTIES
            xSpx, xEpx = math.floor(float(xS) / xMtrPerPx), math.floor(float(xE) / xMtrPerPx)
            ySpx, yEpx = math.floor(float(yS) / yMtrPerPx), math.floor(float(yE) / yMtrPerPx)
            # ADDED CODE TO AVOID ZERO PIXEL REGIONS
            jCorrect = False
            if xSpx == xEpx:
                jCorrect = True
                cnt += 1
                if xEpx + pxAdd < xPxMax:
                    xEpx = xEpx + pxAdd
                else:
                    xSpx = xSpx - pxAdd
            if ySpx == yEpx:
                jCorrect = True
                cnt += 1
                if yEpx + pxAdd < yPxMax:
                    yEpx = yEpx + pxAdd
                else:
                    ySpx = ySpx - pxAdd

            if jCorrect:
                if cnt == 1:
                    print('Correcting for Region sizes less than pixel resolution...')
                # if i == 0: print('  ' + xyNm)

            # END ADDED CODE
            thisReg = img[ySpx:yEpx, xSpx:xEpx]
            y, x, _ = thisReg.shape
            # WRITE REGION BLOCK
            mapFile.write('.REGION,' + regNm + '\n')
            mapFile.write('.REGIONSIZE\n')
            mapFile.write(','.join([xS, xE, yS, yE, zS, zE]) + '\n')
            mapFile.write('.RESOLUTION\n')
            mapFile.write(str(x) + ',' + str(y) + '\n')
            # EXPLICIT
            jTruth = np.all(thisReg[:, :, :3] == blk, axis=2)
            mArrnp = np.where(jTruth, blkP, whtP)  # is True
            mArrFlt = mArrnp.flatten()
            mArr = [m.split(delim) for m in mArrFlt]
            # write fields
            for j, f in enumerate(fields):
                fStr = delim.join([p[j] for p in mArr])
                mapFile.write('.FIELD,' + f + '\n')
                mapFile.write(fStr + '\n')

    if cnt > 0: print('  ' + str(cnt) + ' total correction(s)')

    mapFile.close()


def floPCBimages(floPCBpath, outputFldr):
    imgPaths = []
    tmpfldr = os.path.abspath(os.path.join(floPCBpath, os.pardir))
    with zipfile.ZipFile(floPCBpath, 'r') as zip_ref:
        zip_ref.extractall(tmpfldr)
    fldr = [f for f in os.listdir(tmpfldr) if 'odbImport' in f and not os.path.isfile(f)]
    if len(fldr) > 0:
        imgFldr = os.path.join(tmpfldr, fldr[0] + r'\PDProject')
        imgFiles = [i for i in os.listdir(imgFldr) if i.endswith('PNG')]
        for i in imgFiles:
            imgPath = os.path.join(outputFldr, i)
            imgPaths.append(imgPath)
            shutil.copyfile(src=os.path.join(imgFldr, i), dst=imgPath)
    return imgPaths


def removeCutouts(slnFldr, projNm, cutoutNms):
    fsPath = os.path.join(slnFldr, 'fsRemoveCutout.xml')
    geomPath = os.path.join(slnFldr, 'geometry_model.csv')
    if os.path.exists(geomPath):
        os.rename(geomPath, os.path.join(slnFldr, 'geometry_modelWCutout.csv'))

    fsX = fs.newfs()
    fs.change_solution_directory(fsXML=fsX, directory_path=slnFldr)
    fs.project_load(fsXML=fsX, project_name=projNm)
    fs.change_results_mode(fsXML=fsX, is_results_mode=False)
    for nM in cutoutNms:
        fs.find(fsXML=fsX, geometryName=nM, geometryType='hole')
        fs.deleteAfterFind(fsXML=fsX)
    fs.change_results_mode(fsXML=fsX, is_results_mode=True)
    fs.export_legacy_tables(fsXML=fsX, directory=slnFldr, geometry_model=True)
    fs.quitft(fsXML=fsX)
    fs.savefs(fsXML=fsX, fsPath=fsPath)
    fs.fsplay(fsPath=fsPath, fltBin=fltBin, silent=True)


def importODB(odbPath, slnFldr, userSuppliedGeomCSV=False):
    fsPath = os.path.join(slnFldr, 'fsODB.xml')
    floPCBpath = os.path.join(slnFldr, 'odbImport.flopcb')
    odbNm = os.path.basename(odbPath).split('.')[0]
    if debug: print(odbNm)
    fsX = fs.newfs()
    fs.change_solution_directory(fsXML=fsX, directory_path=slnFldr)
    fs.project_save_as(fsXML=fsX, project_name=odbNm)

    # EDA BRIDGE
    fs.launch(fsXML=fsX, launchType='eda')
    fs.project_save(fsXML=fsX)
    fs.import_odb(fsXML=fsX, odbPath=odbPath, asExternal=True)
    # filtering doesn't seem to work.
    fs.filterEDA(fsXML=fsX, side_length_less_than=2, height_less_than=2, power_less_than=0, power_density_less_than=0)
    fs.edaModifyBoardName(fsXML=fsX, nM=odbNm, asExternal=True)
    fs.edaSaveAs(fsXML=fsX, path=floPCBpath)
    if not userSuppliedGeomCSV:
        fs.edaTransfer(fsXML=fsX)
        fs.edaQuit(fsXML=fsX)

        fs.export_legacy_tables(fsXML=fsX, directory=slnFldr, geometry_model=True)
        if deleteInternalCutouts: fs.project_save(fsXML=fsX)
    fs.quitft(fsXML=fsX)
    fs.savefs(fsXML=fsX, fsPath=fsPath)
    fs.fsplay(fsPath=fsPath, fltBin=fltBin, silent=True)

    return floPCBpath, odbNm


def getEDABrdLayerGeom(geometryCSVPath, pcbNm=''):
    gHdr, gRows = _csvToList(csvPath=geometryCSVPath, hasHeader=True)
    # GET UNIT MULTIPLICATION FACTOR FOR CONVERSION TO meter
    lngthUnit = gHdr[3].split('(')[1][:-1]
    units = ['m', 'cm', 'mm', 'micron', 'in', 'ft', 'U', 'mil']
    mFacs = [1.0, 0.01, 0.001, 1e-6, 0.0254, 0.3048, 0.04445, 2.54e-05]
    mFac = mFacs[units.index(lngthUnit)]
    if matMapUnits: mFac = 1.0
    # GET EDA BOARD INDEX
    EDABrdlnIndexes = [i for i, b in enumerate(gRows) if b[1] == 'EDA Board']
    if len(EDABrdlnIndexes) > 0:
        brdNms = [gRows[index][0] for index in EDABrdlnIndexes]
        if pcbNm in brdNms:
            EDABrdIndex = EDABrdlnIndexes[brdNms.index(pcbNm)]
        else:
            EDABrdIndex = EDABrdlnIndexes[0]
    else:  # No EDA Boards found
        return []

    # PROCESS EDA BOARD
    # GET All EDA BOARD RELATED GEOMETRY
    brdLevel = gRows[EDABrdIndex][2]
    thisLevel = brdLevel
    lnIndex = EDABrdIndex
    edaBrdGeom = []
    while thisLevel >= brdLevel and lnIndex < len(gRows) - 1:
        edaBrdGeom.append(gRows[lnIndex])
        lnIndex += 1

    # PARSE EDA BOARD DATA
    # brdTypes = ['EDA Board', 'EDA Layer Assembly', 'EDA Metallic Layer', 'EDA Dielectric Layer',
    #             'EDA Electrical Vias Assembly']
    brdStack = []
    brdTypes = ['EDA Board', 'EDA Metallic Layer', 'EDA Dielectric Layer']
    typeIndexes = [[i, ln[1]] for i, ln in enumerate(edaBrdGeom) if ln[1] in brdTypes]
    edaBrdNm = edaBrdGeom[typeIndexes[0][0]][0]
    # GET REGIONS: [[name, x Start (m), x End (m), y Start (m), y End (m)]]
    brdRegionXY, brdCutoutXY = [], []
    for thisIndex in range(typeIndexes[0][0], typeIndexes[1][0]):
        thisLn = edaBrdGeom[thisIndex]
        if thisLn[1] == 'Cuboid':
            xStrt, yStrt = float(thisLn[3]), float(thisLn[4])
            xSz, ySz = float(thisLn[6]), float(thisLn[7])
            xEnd, yEnd = xStrt + xSz, yStrt + ySz
            brdRegionXY.append([thisLn[0]] + [str(x * mFac) for x in [xStrt, xEnd, yStrt, yEnd]])
        if thisLn[1] == 'Hole':
            xStrt, yStrt = float(thisLn[3]), float(thisLn[4])
            xSz, ySz = float(thisLn[6]), float(thisLn[7])
            xEnd, yEnd = xStrt + xSz, yStrt + ySz
            brdCutoutXY.append([thisLn[0]] + [str(x * mFac) for x in [xStrt, xEnd, yStrt, yEnd]])

    # GET BOARD STACKUP: [[name,type, z Start (m), z End (m)]]
    brdStack = [edaBrdGeom[k[0]][0:2] + [str(float(edaBrdGeom[k[0]][5]) * mFac)] + [
        str((float(edaBrdGeom[k[0]][5]) + float(edaBrdGeom[k[0]][8])) * mFac)]
                for k in typeIndexes[1:]]

    return edaBrdNm, brdRegionXY, brdCutoutXY, brdStack, lngthUnit


def _pixelUnits(xyRegions, imgPath):
    xMin, xMax = min([float(x[1]) for x in xyRegions]), max([float(x[2]) for x in xyRegions])
    yMin, yMax = min([float(x[3]) for x in xyRegions]), max([float(x[4]) for x in xyRegions])
    xSzBrd, ySzBrd = xMax - xMin, yMax - yMin
    imgNp = cv2.imread(imgPath)
    yPx, xPx, _ = imgNp.shape
    return xSzBrd / xPx, ySzBrd / yPx, yPx, xPx


def _csvToList(csvPath, hasHeader=True):
    import csv
    hdr = []
    with open(csvPath, 'r') as csvfile:
        # creating a csv reader object
        csvreader = csv.reader(csvfile)
        if hasHeader: hdr = next(csvreader)
        rows = [r for r in csvreader if len(r) > 0]
    return hdr, rows


def _cutoutIsConnected(cutoutA, cutoutB, bdryTest=False):
    nMA, xsA, xeA, ysA, yeA = cutoutA
    nMB, xsB, xeB, ysB, yeB = cutoutB
    if bdryTest:
        if abs(xsA - xsB) <= geomTol or abs(xeA - xeB) <= geomTol or abs(ysA - ysB) <= geomTol or abs(
                yeA - yeB) <= geomTol:
            return True
        else:
            return False
    else:
        xTestA = [xsB <= xsA + geomTol <= xeB,
                  xsB <= xsA - geomTol <= xeB,
                  xsB <= xeA + geomTol <= xeB,
                  xsB <= xeA - geomTol <= xeB]
        yTestA = [ysB <= ysA + geomTol <= yeB,
                  ysB <= ysA - geomTol <= yeB,
                  ysB <= yeA + geomTol <= yeB,
                  ysB <= yeA - geomTol <= yeB]
        xTestB = [xsA <= xsB + geomTol <= xeA,
                  xsA <= xsB - geomTol <= xeA,
                  xsA <= xeB + geomTol <= xeA,
                  xsA <= xeB - geomTol <= xeA]
        yTestB = [ysA <= ysB + geomTol <= yeA,
                  ysA <= ysB - geomTol <= yeA,
                  ysA <= yeB + geomTol <= yeA,
                  ysA <= yeB - geomTol <= yeA]

        if True in xTestA and True in yTestA:
            return True
        elif True in xTestB and True in yTestB:
            return True
        else:
            return False


def _sortedList(cutoutA, cutOuts):
    nMb, xsb, xeb, ysb, yeb = cutoutA
    xcb, ycb = 0.5 * (xsb + xeb), 0.5 * (ysb + yeb)
    cutOutsPlus = []
    for c in cutOuts:
        nMi, xsi, xei, ysi, yei = c
        xci, yci = 0.5 * (xsi + xei), 0.5 * (ysi + yei)
        dist = math.sqrt((xcb - xci) ** 2 + (ycb - yci) ** 2)
        cutOutsPlus.append(c + [dist])
    cutOutsPlusSorted = sorted(cutOutsPlus, key=lambda x: x[-1])

    return [c[:-1] for c in cutOutsPlusSorted]


def _repTime(strtTime, msg):
    nowTime = time.time()
    deltaTime = round(nowTime - strtTime, 2)
    if deltaTime < 60:
        print(msg + ' Time [sec]: ', str(deltaTime))
    else:
        deltaTime = round(deltaTime / 60, 2)
        print(msg + ' Time [min]: ', str(deltaTime))
    return nowTime


def help_msg():
    oMsg = ["  CONVERTS ODB++ TO SIMCENTER FLOTHERM MATERIAL MAP FILE",
            "           -i [complete path to ODB++ ARCHIVE FILE]",
            "           -g [OPTIONAL][complete path to Simcenter Flotherm geometry_model.csv FILE]",
            "    TIP: use quotes when specifying the FILE/FOLDER paths"]
    for jmsg in oMsg: print(jmsg)


def arg_process():
    # odbTofltMaterialMap.py -i path to ODB++ File -g [optional] path to geometry_model.csv
    odbPath, geomPath = '', ''
    if len(sys.argv) == 1:
        help_msg()
        quit()
    elif '?' in sys.argv[1]:
        help_msg()
        quit()
    else:
        arg_in = sys.argv[1:]
        for index, i in enumerate(arg_in):
            if i == '-i':
                odbPath = arg_in[index + 1]
            if i == '-g':
                geomPath = arg_in[index + 1]

    return odbPath, geomPath


if __name__ == '__main__':
    main()
